package ru.dbotthepony.mc.otm.core.math

import com.mojang.blaze3d.systems.RenderSystem
import net.minecraft.ChatFormatting
import kotlin.math.roundToInt

data class RGBAColor(val red: Float, val green: Float, val blue: Float, val alpha: Float = 1f) {
	constructor(r: Int, g: Int, b: Int) : this((r / 255f), (g / 255f), (b / 255f), 1f)
	constructor(r: Int, g: Int, b: Int, a: Int) : this((r / 255f), (g / 255f), (b / 255f), (a / 255f))
	constructor(r: Int, g: Int, b: Int, a: Float) : this((r / 255f), (g / 255f), (b / 255f), a)

	constructor(color: Long) : this(
		(((color and -0x1000000) ushr 24) / 255f),
		(((color and 0xFF0000) ushr 16) / 255f),
		(((color and 0xFF00) ushr 8) / 255f),
		(((color and 0xFF)) / 255f)
	)

	fun toInt(): Int {
		val r = (this.red * 255).roundToInt()
		val g = (this.green * 255).roundToInt()
		val b = (this.blue * 255).roundToInt()
		return (r shl 16) or (g shl 8) or b
	}

	fun toRGBA(): Int {
		val r = (this.red * 255).roundToInt()
		val g = (this.green * 255).roundToInt()
		val b = (this.blue * 255).roundToInt()
		val a = (this.alpha * 255).roundToInt()
		return (r shl 24) or (g shl 16) or (b shl 8) or a
	}

	fun toARGB(): Int {
		val r = (this.red * 255).roundToInt()
		val g = (this.green * 255).roundToInt()
		val b = (this.blue * 255).roundToInt()
		val a = (this.alpha * 255).roundToInt()
		return (a shl 24) or (r shl 16) or (g shl 8) or b
	}

	fun toBGRA(): Int {
		val r = (this.red * 255).roundToInt()
		val g = (this.green * 255).roundToInt()
		val b = (this.blue * 255).roundToInt()
		val a = (this.alpha * 255).roundToInt()
		return (r shl 24) or (g shl 16) or (r shl 8) or a
	}

	fun toIntInv(): Int {
		val r = (this.red * 255).roundToInt()
		val g = (this.green * 255).roundToInt()
		val b = (this.blue * 255).roundToInt()
		return (b shl 16) or (g shl 8) or r
	}

	fun setSystemColor() {
		setShaderColor()
		setDrawColor()
	}

	fun setShaderColor() {
		RenderSystem.setShaderColor(red, green, blue, alpha)
	}

	fun setDrawColor() {
		ru.dbotthepony.mc.otm.client.render.setDrawColor(this)
	}

	fun linearInterpolation(t: Float, other: RGBAColor): RGBAColor {
		return RGBAColor(
			linearInterpolation(t, red, other.red),
			linearInterpolation(t, green, other.green),
			linearInterpolation(t, blue, other.blue),
			linearInterpolation(t, alpha, other.alpha),
		)
	}

	val isFullyTransparent get() = alpha <= 0f

	companion object {
		@JvmField val TRANSPARENT_BLACK = RGBAColor(0f, 0f, 0f, 0f)
		@JvmField val TRANSPARENT_WHITE = RGBAColor(1f, 1f, 1f, 0f)

		@JvmField val BLACK = RGBAColor(0f, 0f, 0f, 1f)
		@JvmField val WHITE = RGBAColor(1f, 1f, 1f, 1f)
		@JvmField val RED = RGBAColor(1f, 0f, 0f)
		@JvmField val GREEN = RGBAColor(0f, 1f, 0f, 1f)
		@JvmField val LIGHT_GREEN = RGBAColor(136, 255, 124)
		@JvmField val SLATE_GRAY = RGBAColor(64, 64, 64)
		@JvmField val GRAY = RGBAColor(0x2C2C2CFFL)

		@JvmField val DARK_BLUE = rgb(ChatFormatting.DARK_BLUE.color!!)
		@JvmField val DARK_GREEN = rgb(ChatFormatting.DARK_GREEN.color!!)
		@JvmField val DARK_AQUA = rgb(ChatFormatting.DARK_AQUA.color!!)
		@JvmField val DARK_RED = rgb(ChatFormatting.DARK_RED.color!!)
		@JvmField val DARK_PURPLE = rgb(ChatFormatting.DARK_PURPLE.color!!)
		@JvmField val GOLD = rgb(ChatFormatting.GOLD.color!!)
		@JvmField val DARK_GRAY = rgb(ChatFormatting.DARK_GRAY.color!!)
		@JvmField val BLUE = rgb(ChatFormatting.BLUE.color!!)
		@JvmField val AQUA = rgb(ChatFormatting.AQUA.color!!)
		@JvmField val LIGHT_PURPLE = rgb(ChatFormatting.LIGHT_PURPLE.color!!)
		@JvmField val YELLOW = rgb(ChatFormatting.YELLOW.color!!)

		@JvmField val LOW_POWER = RGBAColor(173, 41, 41)
		@JvmField val FULL_POWER = RGBAColor(255, 242, 40)
		@JvmField val LOW_MATTER = RGBAColor(0, 24, 148)
		@JvmField val FULL_MATTER = RGBAColor(72, 90, 255)
		@JvmField val LOW_PATTERNS = RGBAColor(44, 104, 57)
		@JvmField val FULL_PATTERNS = RGBAColor(65, 255, 87)

		fun inv(color: Int): RGBAColor {
			val r = (color and -0x1000000 ushr 24) / 255f
			val g = (color and 0xFF0000 ushr 16) / 255f
			val b = (color and 0xFF00 ushr 8) / 255f
			val a = (color and 0xFF) / 255f
			return RGBAColor(a, b, g, r)
		}

		fun rgb(color: Int): RGBAColor {
			val r = (color and 0xFF0000 ushr 16) / 255f
			val g = (color and 0xFF00 ushr 8) / 255f
			val b = (color and 0xFF) / 255f
			return RGBAColor(r, g, b)
		}

		fun argb(color: Int): RGBAColor {
			val a = (color and -0x1000000 ushr 24) / 255f
			val r = (color and 0xFF0000 ushr 16) / 255f
			val g = (color and 0xFF00 ushr 8) / 255f
			val b = (color and 0xFF) / 255f
			return RGBAColor(r, g, b, a)
		}
	}
}

fun linearInterpolation(t: Float, a: RGBAColor, b: RGBAColor): RGBAColor {
	return a.linearInterpolation(t, b)
}
